import tkinter as tk
from tkinter import font  as tkfont
from tkinter import filedialog
from tkinter import ttk
from tkinter import *


class AdujstParamInsideAlgo(tk.Frame):

  def __init__(self, parent, controller):
  
    tk.Frame.__init__(self, parent)
    self.controller = controller
    label = tk.Label(self, text="Advanced Parameter adjustment", font=controller.title_font)
    label.pack(side="top", fill="x", pady=10)
    
    tk.Label(self, text="Recalculate background using this number of images: (default is 60)", font=("Helvetica", 10)).pack(side="top", fill="x", pady=10)
    nbImagesForBackgroundCalculation = tk.Entry(self)
    nbImagesForBackgroundCalculation.pack()
    tk.Button(self, text="Recalculate", command=lambda: controller.calculateBackground(controller, nbImagesForBackgroundCalculation.get())).pack()
    
    tk.Label(self, text="", pady=5).pack()
    firstFrameParamAdjust = IntVar()
    adjustOnWholeVideo    = IntVar()
    Checkbutton(self, text="Choose the first frame for parameter adjustment (for both bouts detection and tracking)", variable=firstFrameParamAdjust).pack()
    Checkbutton(self, text="I want to adjust parameters over the entire video, not only on 500 frames at a time.", variable=adjustOnWholeVideo).pack()
    tk.Label(self, text="", pady=5).pack()
    
    tk.Button(self, text="Adjust Bouts Detection", command=lambda: controller.detectBouts(controller, "0", firstFrameParamAdjust.get(), adjustOnWholeVideo.get())).pack()
    tk.Label(self, text="The aim here is to adjust parameters in order for the red dot on the top left of the image to appear when and only when movement is occurring.", font=("Helvetica", 10)).pack(side="top", fill="x", pady=10)
    tk.Label(self, text="WARNING: if you don't want ZebraZoom to detect bouts, don't click on the button above.", font=("Helvetica", 10)).pack(side="top", fill="x", pady=0)
    tk.Label(self, text="", font=("Helvetica", 0)).pack(side="top", fill="x", pady=0)
    
    tk.Button(self, text="Adjust Tracking", command=lambda: controller.adjustHeadEmbededTracking(controller, "0", firstFrameParamAdjust.get(), adjustOnWholeVideo.get())).pack()
    
    tk.Label(self, text="WARNING: only click the button above if you've tried to track without adjusting these parameters first. Trying to adjust these could make the tracking worse.", font=("Helvetica", 10)).pack(side="top", fill="x", pady=0)
    tk.Label(self, text='Warning: for some of the "overwrite" parameters, you will need to change the initial value for the "overwrite" to take effect.', font=("Helvetica", 10)).pack(side="top", fill="x", pady=0)
    tk.Label(self, text='', font=("Helvetica", 10)).pack(side="top", fill="x", pady=0)
    
    tk.Button(self, text="Next", command=lambda: controller.show_frame("FinishConfig")).pack()
    

class AdujstParamInsideAlgoFreelySwim(tk.Frame):

  def __init__(self, parent, controller):
  
    tk.Frame.__init__(self, parent)
    self.controller = controller
    label = tk.Label(self, text="Advanced Parameter adjustment", font=controller.title_font)
    label.pack(side="top", fill="x", pady=10)
    
    tk.Label(self, text="Well number used to adjust parameters (leave blank to get the default value of 0)", font=("Helvetica", 10)).pack(side="top", fill="x", pady=10)
    wellNumber = tk.Entry(self)
    wellNumber.pack()
    
    tk.Label(self, text="Recalculate background using this number of images: (default is 60)", font=("Helvetica", 10)).pack(side="top", fill="x", pady=10)
    nbImagesForBackgroundCalculation = tk.Entry(self)
    nbImagesForBackgroundCalculation.pack()
    tk.Button(self, text="Recalculate", command=lambda: controller.calculateBackgroundFreelySwim(controller, nbImagesForBackgroundCalculation.get())).pack()
    
    tk.Label(self, text="", pady=5).pack()
    firstFrameParamAdjust = IntVar()
    adjustOnWholeVideo    = IntVar()
    Checkbutton(self, text="Choose the first frame for parameter adjustment (for both bouts detection and tracking)", variable=firstFrameParamAdjust).pack()
    Checkbutton(self, text="I want to adjust parameters over the entire video, not only on 500 frames at a time.", variable=adjustOnWholeVideo).pack()
    tk.Label(self, text="", pady=5).pack()
    
    tk.Button(self, text="Adjust Bouts Detection", command=lambda: controller.detectBouts(controller, wellNumber.get(), firstFrameParamAdjust.get(), adjustOnWholeVideo.get())).pack()
    tk.Label(self, text="The aim here is to adjust parameters in order for the red dot on the top left of the image to appear when and only when movement is occurring.", font=("Helvetica", 10)).pack(side="top", fill="x", pady=10)
    tk.Label(self, text="WARNING: if you don't want ZebraZoom to detect bouts, don't click on the button above.", font=("Helvetica", 10)).pack(side="top", fill="x", pady=0)
    tk.Label(self, text="", font=("Helvetica", 0)).pack(side="top", fill="x", pady=0)
    
    tk.Button(self, text="Adjust Tracking", command=lambda: controller.adjustFreelySwimTracking(controller, wellNumber.get(), firstFrameParamAdjust.get(), adjustOnWholeVideo.get())).pack()
    tk.Label(self, text="WARNING: only click the button above if you've tried to track without adjusting these parameters first. Trying to adjust these could make the tracking worse.", font=("Helvetica", 10)).pack(side="top", fill="x", pady=0)
    tk.Label(self, text='', font=("Helvetica", 10)).pack(side="top", fill="x", pady=0)
    
    tk.Button(self, text="Next", command=lambda: controller.show_frame("FinishConfig")).pack()


class AdujstParamInsideAlgoFreelySwimAutomaticParameters(tk.Frame):

  def __init__(self, parent, controller):
  
    tk.Frame.__init__(self, parent)
    self.controller = controller
    label = tk.Label(self, text="Fish tail tracking parameters adjustment", font=controller.title_font, pady=30)
    label.grid(row=0,column=0, columnspan = 2)
    
    tk.Label(self, text="Well number used to adjust parameters (leave blank to get the default value of 0)", font=("Helvetica", 10)).grid(row=1,column=0)
    wellNumber = tk.Entry(self)
    wellNumber.grid(row=2,column=0)
    
    tk.Label(self, text="Recalculate background using this number of images: (default is 60)", font=("Helvetica", 10)).grid(row=1,column=1)
    nbImagesForBackgroundCalculation = tk.Entry(self)
    nbImagesForBackgroundCalculation.grid(row=2,column=1)
    tk.Button(self, text="Recalculate", command=lambda: controller.calculateBackgroundFreelySwim(controller, nbImagesForBackgroundCalculation.get(), False, True)).grid(row=3,column=1)
    
    # tk.Label(self, text="", pady=5).grid(row=1,column=0)
    firstFrameParamAdjust = IntVar()
    adjustOnWholeVideo    = IntVar()
    Checkbutton(self, text="Choose the first frame for parameter adjustment (for both bouts detection and tracking)", variable=firstFrameParamAdjust).grid(row=3,column=0)
    Checkbutton(self, text="I want to adjust parameters over the entire video, not only on 500 frames at a time.", variable=adjustOnWholeVideo).grid(row=4,column=0)
    # tk.Label(self, text="", pady=5).pack()
    
    tk.Label(self, text="", pady=8).grid(row=6,column=1)
    tk.Button(self, text="Adjust Tracking", command=lambda: controller.adjustFreelySwimTrackingAutomaticParameters(controller, wellNumber.get(), firstFrameParamAdjust.get(), adjustOnWholeVideo.get())).grid(row=7,column=0, columnspan = 2)
    tk.Label(self, text="").grid(row=8,column=1)
    tk.Label(self, text="The tracking of ZebraZoom can rely on three different background extraction methods:").grid(row=9,column=0, columnspan = 2)
    tk.Label(self, text="Method 1: background extraction is based on a simple threshold on pixel intensity. This method is the fastest.").grid(row=10,column=0, columnspan = 2)
    tk.Label(self, text="Method 2: the background extraction threshold is automatically chosen in order for the fish body area to be close to a predefined area. This method is slower but often more accurate.").grid(row=11,column=0, columnspan = 2)
    tk.Label(self, text="Method 3: the background extraction threshold is automatically chosen on a ROI in order for the fish body area to be close to a predefined area. This method is the slowest but often the most accurate.").grid(row=12,column=0, columnspan = 2)
    tk.Label(self, text="It is usually advise to choose the method 3, but there are many circumstances in which method 1 or 2 are better.").grid(row=13,column=0, columnspan = 2)
    tk.Label(self, text="The 'Adjust Tracking' method above will allow you to choose which method you want to use and to adjust parameters related to this method.").grid(row=14,column=0, columnspan = 2)
    tk.Label(self, text="WARNING: only click the button above if you've tried to track without adjusting these parameters first. Trying to adjust these could make the tracking worse.", font=("Helvetica", 10)).grid(row=15,column=0, columnspan = 2)
    
    tk.Label(self, text="", font=("Helvetica", 10), pady=5).grid(row=16,column=0, columnspan = 2)
    tk.Button(self, text="Save New Configuration File", command=lambda: controller.show_frame("FinishConfig")).grid(row=17,column=0, columnspan = 2)
    tk.Label(self, text="", font=("Helvetica", 10), pady=5).grid(row=18,column=0, columnspan = 2)

    tk.Button(self, text="Go to the start page", bg="light cyan", command=lambda: controller.show_frame("StartPage")).grid(row=19,column=0, columnspan = 2)


class AdujstBoutDetectionOnly(tk.Frame):

  def __init__(self, parent, controller):
  
    tk.Frame.__init__(self, parent)
    self.controller = controller
    label = tk.Label(self, text="Bout detection configuration file parameters adjustments", font=controller.title_font)
    label.pack(side="top", fill="x", pady=10)
    
    tk.Label(self, text="Well number used to adjust parameters (leave blank to get the default value of 0)", font=("Helvetica", 10)).pack(side="top", fill="x", pady=10)
    wellNumber = tk.Entry(self)
    wellNumber.pack()
    
    tk.Label(self, text="Recalculate background using this number of images: (default is 60)", font=("Helvetica", 10)).pack(side="top", fill="x", pady=10)
    nbImagesForBackgroundCalculation = tk.Entry(self)
    nbImagesForBackgroundCalculation.pack()
    tk.Button(self, text="Recalculate", command=lambda: controller.calculateBackgroundFreelySwim(controller, nbImagesForBackgroundCalculation.get(), False, False, True)).pack()
    
    tk.Label(self, text="", pady=5).pack()
    firstFrameParamAdjust = IntVar()
    adjustOnWholeVideo    = IntVar()
    Checkbutton(self, text="Choose the first frame for parameter adjustment (for both bouts detection and tracking)", variable=firstFrameParamAdjust).pack()
    Checkbutton(self, text="I want to adjust parameters over the entire video, not only on 500 frames at a time.", variable=adjustOnWholeVideo).pack()
    tk.Label(self, text="", pady=5).pack()
    
    tk.Button(self, text="Adjust Bouts Detection", command=lambda: controller.detectBouts(controller, wellNumber.get(), firstFrameParamAdjust.get(), adjustOnWholeVideo.get())).pack()
    tk.Label(self, text="The aim here is to adjust parameters in order for the red dot on the top left of the image to appear when and only when movement is occurring.", font=("Helvetica", 10)).pack(side="top", fill="x", pady=10)
    
    tk.Label(self, text="", font=("Helvetica", 0)).pack(side="top", fill="x", pady=0)
    tk.Label(self, text="Important: Bouts Merging:", font=("Helvetica", 0)).pack(side="top", fill="x", pady=0)
    fillGapFrameNb = tk.Entry(self)
    fillGapFrameNb.pack()
    tk.Button(self, text="With the box above, update the 'fillGapFrameNb' parameter that controls the distance (in number frames) under which two subsquent bouts are merged into one.", command=lambda: controller.updateFillGapFrameNb(fillGapFrameNb.get())).pack()    
    tk.Label(self, text="", font=("Helvetica", 0)).pack(side="top", fill="x", pady=0)
    
    tk.Button(self, text="Next", command=lambda: controller.show_frame("FinishConfig")).pack()
    
    tk.Button(self, text="Go to the start page", bg="light cyan", command=lambda: controller.show_frame("StartPage")).pack()


import numpy as np
import cv2
import math

def rollingMedianFilter(array, window):
  array2 = np.convolve(array, np.ones(window), 'same') / window
  array2[:window-1] = array[:window-1]
  array2[-window+1:] = array[-window+1:]
  array = array2
  return array

def postProcessMultipleTrajectories(trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals, trackingProbabilityOfGoodDetection, hyperparameters, wellPositions):
  
  maxDistanceAuthorized = hyperparameters["postProcessMaxDistanceAuthorized"]
  maxDisapearanceFrames = hyperparameters["postProcessMaxDisapearanceFrames"]
  
  # Removing all points for which the sum of all pixels of the inverted image (the "probability of good detection") is too low
  if hyperparameters["postProcessRemoveLowProbabilityDetection"]:
    for animalId in range(0, len(trackingHeadTailAllAnimals)):
      probabilityOfGoodDetection = trackingProbabilityOfGoodDetection[animalId, :]
      if hyperparameters["postProcessLowProbabilityDetectionPercentOfMaximum"] == 0:
        toRemove   = (probabilityOfGoodDetection - np.mean(probabilityOfGoodDetection) < -hyperparameters["postProcessLowProbabilityDetectionThreshold"] * np.std(probabilityOfGoodDetection))
      else:
        toRemove   = (probabilityOfGoodDetection < np.max(probabilityOfGoodDetection) * hyperparameters["postProcessLowProbabilityDetectionPercentOfMaximum"])
      print("Well ", animalId, ": Number of elements to remove:", np.sum(toRemove), " out of ", len(toRemove), "elements")
      trackingHeadTailAllAnimals[animalId, toRemove, 0, 0] = 0
      trackingHeadTailAllAnimals[animalId, toRemove, 0, 1] = 0
  
  # Removing all points that are too close to the borders
  if hyperparameters["postProcessRemovePointsOnBordersMargin"]:
    borderMargin = hyperparameters["postProcessRemovePointsOnBordersMargin"]
    for animalId in range(0, len(trackingHeadTailAllAnimals)):
      for frameNumber in range(0, len(trackingHeadTailAllAnimals[animalId])):
        xHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
        yHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
        if (xHead <= borderMargin) or (yHead <= borderMargin) or (xHead >= wellPositions[animalId]["lengthX"] - borderMargin - 1) or (yHead >= wellPositions[animalId]["lengthY"] - borderMargin - 1):
          trackingHeadTailAllAnimals[animalId][frameNumber][0][0] = 0
          trackingHeadTailAllAnimals[animalId][frameNumber][0][1] = 0
  
  # Removing all points that are deviating too much from the main trajectory
  if hyperparameters["postProcessRemovePointsAwayFromMainTrajectory"]:
    for animalId in range(0, len(trackingHeadTailAllAnimals)):
      xHeadPositions = trackingHeadTailAllAnimals[animalId, :, 0, 0]
      yHeadPositions = trackingHeadTailAllAnimals[animalId, :, 0, 1]
      xHeadPositionsRollingMedian = rollingMedianFilter(xHeadPositions, 11)
      yHeadPositionsRollingMedian = rollingMedianFilter(yHeadPositions, 11)
      distance = np.sqrt((xHeadPositions - xHeadPositionsRollingMedian) ** 2 + (yHeadPositions - yHeadPositionsRollingMedian) ** 2)
      distanceRollingMedian = rollingMedianFilter(distance, 5)
      normalizedDistance = np.array([1 for x in range(0,len(xHeadPositions))])
      normalizedDistance = distance / distanceRollingMedian
      normalizedDistance = np.nan_to_num(normalizedDistance, nan=1)
      toRemove   = normalizedDistance - np.mean(normalizedDistance) > hyperparameters["postProcessRemovePointsAwayFromMainTrajectoryThreshold"] * np.std(normalizedDistance)
      trackingHeadTailAllAnimals[animalId, toRemove, 0, 0] = 0
      trackingHeadTailAllAnimals[animalId, toRemove, 0, 1] = 0
  
  currentlyZero  = False
  zeroFrameStart = 0
  for animalId in range(0, len(trackingHeadTailAllAnimals)):
    for frameNumber in range(0, len(trackingHeadTailAllAnimals[animalId])):
      xHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
      yHead = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
      if frameNumber != 0:
        xHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber-1][0][0]
        yHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber-1][0][1]
      else:
        xHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
        yHeadPrev = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
      if ((xHead == 0 and yHead == 0) or (math.sqrt((xHead - xHeadPrev)**2 + (yHead - yHeadPrev)**2) > maxDistanceAuthorized)) and (frameNumber != len(trackingHeadTailAllAnimals[animalId]) - 1):
        if not(currentlyZero):
          zeroFrameStart = frameNumber
        # print("currentlyZero at True: animalId:", animalId, "; frameNumber:", frameNumber)
        currentlyZero = True
      else:
        if currentlyZero:
          if zeroFrameStart >= 1:
            xHeadStart = trackingHeadTailAllAnimals[animalId][zeroFrameStart-1][0][0]
            yHeadStart = trackingHeadTailAllAnimals[animalId][zeroFrameStart-1][0][1]
          else:
            xHeadStart = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
            yHeadStart = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
          xHeadEnd = trackingHeadTailAllAnimals[animalId][frameNumber][0][0]
          yHeadEnd = trackingHeadTailAllAnimals[animalId][frameNumber][0][1]
          if (math.sqrt((xHeadEnd - xHeadStart)**2 + (yHeadEnd - yHeadStart)**2) < maxDistanceAuthorized) or (frameNumber - zeroFrameStart > maxDisapearanceFrames):
            currentlyZero = False
            if (math.sqrt((xHeadEnd - xHeadStart)**2 + (yHeadEnd - yHeadStart)**2) < maxDistanceAuthorized) and not((xHeadEnd == 0) and (yHeadEnd == 0)):
              xStep = (xHeadEnd - xHeadStart) / (frameNumber - zeroFrameStart)
              yStep = (yHeadEnd - yHeadStart) / (frameNumber - zeroFrameStart)
              for frameAtZeroToChange in range(zeroFrameStart, frameNumber):
                trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][0] = xHeadStart + xStep * (frameAtZeroToChange - zeroFrameStart)
                trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][1] = yHeadStart + yStep * (frameAtZeroToChange - zeroFrameStart)    
            else:
              for frameAtZeroToChange in range(zeroFrameStart, frameNumber):
                trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][0] = xHeadStart
                trackingHeadTailAllAnimals[animalId][frameAtZeroToChange][0][1] = yHeadStart
  
  return [trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingEyesAllAnimals]



import zebrazoom

# Creating the dataframe

dataframeOptions = {
  'pathToExcelFile'                   : './experimentOrganizationExcel/',
  'fileExtension'                     : '.xls',
  'resFolder'                         : 'data',
  'nameOfFile'                        : 'example',
  'smoothingFactorDynaParam'          : 0,   # 0.001
  'nbFramesTakenIntoAccount'          : 28,
  'numberOfBendsIncludedForMaxDetect' : -1,
  'minNbBendForBoutDetect'            : 3,
  'defaultZZoutputFolderPath'         : '../ZZoutput/',
  'computeTailAngleParamForCluster'   : False,
  'computeMassCenterParamForCluster'  : False
}

[conditions, genotypes, nbFramesTakenIntoAccount, globParam] = zebrazoom.createDataFrame(dataframeOptions)


# Plotting for the different conditions
nameOfFile = dataframeOptions['nameOfFile']
resFolder  = dataframeOptions['resFolder']
globParam = ['BoutDuration', 'TotalDistance', 'Speed', 'NumberOfOscillations', 'meanTBF', 'maxAmplitude']

zebrazoom.populationComparaison(nameOfFile, resFolder, globParam, conditions, genotypes, 'resultsKinematic')




from zebrazoom.code.trackingFolder.tracking import tracking
from zebrazoom.code.extractParameters import extractParameters
from zebrazoom.code.trackingFolder.headTrackingHeadingCalculationFolder.headTrackingHeadingCalculation import headTrackingHeadingCalculation
from zebrazoom.code.trackingFolder.tailTracking import tailTracking
from zebrazoom.code.trackingFolder.eyeTracking.eyeTracking import eyeTracking
from zebrazoom.code.trackingFolder.postProcessMultipleTrajectories import postProcessMultipleTrajectories
from zebrazoom.code.trackingFolder.getImages import getImages
from zebrazoom.code.trackingFolder.debugTracking import debugTracking
from zebrazoom.code.popUpAlgoFollow import prepend
from zebrazoom.code.adjustHyperparameters import adjustHyperparameters, getFreelySwimTrackingParamsForHyperParamAdjusts
import multiprocessing as mp
from multiprocessing import Process
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import numpy as np
import math

def fasterMultiprocessing(videoPath, background, wellPositions, output, hyperparameters, videoName):
  
  cap = zzVideoReading.VideoCapture(videoPath)
  if (cap.isOpened()== False): 
    print("Error opening video stream or file")
  frame_width  = int(cap.get(3))
  frame_height = int(cap.get(4))
  firstFrame = hyperparameters["firstFrame"]
  lastFrame  = hyperparameters["lastFrame"]
  nbTailPoints = hyperparameters["nbTailPoints"]
  
  trackingHeadTailAllAnimalsList = []
  trackingHeadingAllAnimalsList  = []
  if hyperparameters["eyeTracking"]:
    trackingEyesAllAnimalsList   = []
  else:
    trackingEyesAllAnimals = 0
  trackingDataList               = []
  
  if not(hyperparameters["nbAnimalsPerWell"] > 1) and not(hyperparameters["headEmbeded"]) and (hyperparameters["findHeadPositionByUserInput"] == 0) and (hyperparameters["takeTheHeadClosestToTheCenter"] == 0):
    trackingProbabilityOfGoodDetectionList = []
  else:
    trackingProbabilityOfGoodDetectionList = 0
  
  for wellNumber in range(0, hyperparameters["nbWells"]):
    trackingHeadTailAllAnimalsList.append(np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1, nbTailPoints, 2)))
    trackingHeadingAllAnimalsList.append(np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1)))
    if hyperparameters["eyeTracking"]:
      trackingEyesAllAnimalsList.append(np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1, 8)))
    if not(hyperparameters["nbAnimalsPerWell"] > 1) and not(hyperparameters["headEmbeded"]) and (hyperparameters["findHeadPositionByUserInput"] == 0) and (hyperparameters["takeTheHeadClosestToTheCenter"] == 0):
      trackingProbabilityOfGoodDetectionList.append(np.zeros((hyperparameters["nbAnimalsPerWell"], lastFrame-firstFrame+1)))
  
  if hyperparameters["backgroundSubtractorKNN"]:
    fgbg = cv2.createBackgroundSubtractorKNN()
    for i in range(0, min(lastFrame - 1, 500), int(min(lastFrame - 1, 500) / 10)):
      cap.set(1, min(lastFrame - 1, 500) - i)
      ret, frame = cap.read()
      fgmask = fgbg.apply(frame)
    cap.release()
    cap = zzVideoReading.VideoCapture(videoPath)
  
  i = firstFrame
  
  if firstFrame:
    cap.set(1, firstFrame)
  
  while (i < lastFrame + 1):
    
    if (hyperparameters["freqAlgoPosFollow"] != 0) and (i % hyperparameters["freqAlgoPosFollow"] == 0):
      print("Tracking: frame:",i)
      if hyperparameters["popUpAlgoFollow"]:
        prepend("Tracking: frame:" + str(i))
    
    if hyperparameters["debugTracking"]:
      print("frame:",i)
      
    ret, frame = cap.read()
    
    if ret:
    
      if hyperparameters["backgroundSubtractorKNN"]:
        frame = fgbg.apply(frame)
        frame = 255 - frame
      
      for wellNumber in range(0 if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"], hyperparameters["nbWells"] if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"] + 1):
        
        if hyperparameters["nbAnimalsPerWell"] == 1 and not(hyperparameters["forceBlobMethodForHeadTracking"]):
          minPixelDiffForBackExtract = hyperparameters["minPixelDiffForBackExtract"]
          xtop = wellPositions[wellNumber]['topLeftX']
          ytop = wellPositions[wellNumber]['topLeftY']
          lenX = wellPositions[wellNumber]['lengthX']
          lenY = wellPositions[wellNumber]['lengthY']
          if hyperparameters["backgroundSubtractorKNN"]:
            grey = frame
          else:
            grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
          if not(hyperparameters["backgroundSubtractorKNN"]):
            back = background[ytop:ytop+lenY, xtop:xtop+lenX]
            putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
            curFrame[putToWhite] = 255
          else:
            hyperparameters["paramGaussianBlur"] = int(math.sqrt(cv2.countNonZero(255 - curFrame) / hyperparameters["nbAnimalsPerWell"]) / 2) * 2 + 1
          if hyperparameters["paramGaussianBlur"]:
            blur = cv2.GaussianBlur(curFrame, (hyperparameters["paramGaussianBlur"], hyperparameters["paramGaussianBlur"]),0)
          else:
            blur = curFrame
          headPositionFirstFrame = 0
          thresh1 = 0
          thresh2 = 0
          gray    = 0
        else:
          [frame2, gray, thresh1, blur, thresh2, frame2, initialCurFrame, back, xHead, yHead] = getImages(hyperparameters, 0, videoPath, i, background, wellNumber, wellPositions, frame)
          headPositionFirstFrame = 0
        
        # Head tracking and heading calculation
        
        [trackingHeadingAllAnimalsList[wellNumber], trackingHeadTailAllAnimalsList[wellNumber], trackingProbabilityOfGoodDetectionList[wellNumber],lastFirstTheta] = headTrackingHeadingCalculation(hyperparameters, firstFrame, i, blur, thresh1, thresh2, gray, hyperparameters["erodeSize"], frame_width, frame_height, trackingHeadingAllAnimalsList[wellNumber], trackingHeadTailAllAnimalsList[wellNumber], trackingProbabilityOfGoodDetectionList[wellNumber], headPositionFirstFrame, wellPositions[wellNumber]["lengthX"])
        
        # Tail tracking for frame i
        if hyperparameters["trackTail"] == 1 :
          threshForBlackFrames = 0
          thetaDiffAccept = 1.2
          lastFirstTheta = 0
          maxDepth = 0
          tailTipFirstFrame = []
          for animalId in range(0, hyperparameters["nbAnimalsPerWell"]):
            [trackingHeadTailAllAnimalsList[wellNumber], trackingHeadingAllAnimalsList[wellNumber]] = tailTracking(animalId, i, firstFrame, videoPath, frame, hyperparameters, thresh1, nbTailPoints, threshForBlackFrames, thetaDiffAccept, trackingHeadTailAllAnimalsList[wellNumber], trackingHeadingAllAnimalsList[wellNumber], lastFirstTheta, maxDepth, tailTipFirstFrame, initialCurFrame.copy(), back)
        
        # Eye tracking for frame i
        if hyperparameters["eyeTracking"]:
          trackingEyesAllAnimalsList[wellNumber] = eyeTracking(animalId, i, firstFrame, frame, hyperparameters, thresh1, trackingHeadingAllAnimalsList[wellNumber], trackingHeadTailAllAnimalsList[wellNumber], trackingEyesAllAnimalsList[wellNumber])
        
        debugTracking(nbTailPoints, i, firstFrame, trackingHeadTailAllAnimalsList[wellNumber], trackingHeadingAllAnimalsList[wellNumber], blur, hyperparameters)
        
        if hyperparameters["freqAlgoPosFollow"]:
          if i % hyperparameters["freqAlgoPosFollow"] == 0:
            print("Tracking at frame", i)
        
    if hyperparameters["adjustFreelySwimTracking"] == 1:
      [hyperparametersListNames, frameToShow, WINDOW_NAME, organizationTab] = getFreelySwimTrackingParamsForHyperParamAdjusts(nbTailPoints, i, firstFrame, trackingHeadTailAllAnimals, trackingHeadingAllAnimals, frame, frame2, hyperparameters)
      if len(organizationTabCur) == 0:
        organizationTabCur = organizationTab
      [i, hyperparameters, organizationTabCur] = adjustHyperparameters(i, hyperparameters, hyperparametersListNames, frameToShow, WINDOW_NAME, organizationTabCur)
    else:
      i = i + 1
    
  for wellNumber in range(0 if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"], hyperparameters["nbWells"] if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"] + 1):
    
    [trackingHeadingAllAnimalsList[wellNumber], trackingHeadTailAllAnimalsList[wellNumber], trackingEyesAllAnimals] = postProcessMultipleTrajectories(trackingHeadingAllAnimalsList[wellNumber], trackingHeadTailAllAnimalsList[wellNumber], [], trackingProbabilityOfGoodDetectionList[wellNumber], hyperparameters, wellPositions)
    
    trackingDataList.append([trackingHeadTailAllAnimalsList[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0])
  
  wellNumberBeginLoop = 0 if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"]
  for wellNumber in range(wellNumberBeginLoop, hyperparameters["nbWells"] if hyperparameters["onlyTrackThisOneWell"] == -1 else hyperparameters["onlyTrackThisOneWell"] + 1):
    parameters = extractParameters(trackingDataList[wellNumber-wellNumberBeginLoop], wellNumber, hyperparameters, videoPath, wellPositions, background)
    output.append([wellNumber,parameters,[]])
  
  return output


from pathlib import Path
import numpy as np
import tkinter as tk
from tkinter import font  as tkfont
from tkinter import filedialog
from tkinter import ttk
from tkinter import *
import json
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.GUI.getCoordinates import findWellLeft, findWellRight, findHeadCenter, findBodyExtremity
from zebrazoom.code.GUI.automaticallyFindOptimalParameters import automaticallyFindOptimalParameters
import math
from zebrazoom.code.findWells import findWells
from zebrazoom.code.getHyperparameters import getHyperparametersSimple
from zebrazoom.code.getBackground import getBackground
from zebrazoom.code.getImage.getForegroundImage import getForegroundImage
import cvui
import pickle
from zebrazoom.code.vars import getGlobalVariables
from zebrazoom.mainZZ import mainZZ
import json
import os
from zebrazoom.code.resizeImageTooLarge import resizeImageTooLarge
globalVariables = getGlobalVariables()

def getMainArguments(self):
  s            = self.videoToCreateConfigFileFor
  arr          = s.split("/")
  nameWithExt  = arr.pop()
  pathToVideo  = '/'.join(arr) + '/'
  nameWithExtL = nameWithExt.split(".")
  videoExt     = nameWithExtL.pop()
  videoName    = '.'.join(nameWithExtL)
  configFile   = self.configFile
  argv         = []
  return [pathToVideo, videoName, videoExt, configFile, argv]

def chooseVideoToCreateConfigFileFor(self, controller, reloadConfigFile, freelySwimAutomaticParameters=False, boutDetectionsOnly=False):

  if int(reloadConfigFile):
  
    cur_dir_path = os.path.dirname(os.path.realpath(__file__))
    pathconf = Path(cur_dir_path)
    pathconf = pathconf.parent.parent
    pathconf = os.path.join(pathconf, 'configuration/')
    
    if globalVariables["mac"]:
      configFileName =  filedialog.askopenfilename(initialdir = pathconf, title = "Select configuration file")
    else:
      configFileName =  filedialog.askopenfilename(initialdir = pathconf, title = "Select configuration file", filetypes = (("json files","*.json"),("all files","*.*")))
    with open(configFileName) as f:
      self.configFile = json.load(f)

  if globalVariables["mac"]:
    self.videoToCreateConfigFileFor = filedialog.askopenfilename(initialdir = os.path.expanduser("~"),title = "Select video to create config file for")
  else:
    self.videoToCreateConfigFileFor = filedialog.askopenfilename(initialdir = os.path.expanduser("~"),title = "Select video to create config file for",filetypes = (("video","*.*"),("all files","*.*")))
  
  if boutDetectionsOnly:
    controller.calculateBackgroundFreelySwim(controller, 0, False, False, True)
  else:
    if freelySwimAutomaticParameters:
      controller.calculateBackgroundFreelySwim(controller, 0, False, True)
    else:
      controller.show_frame("ChooseGeneralExperiment")

def chooseGeneralExperimentFirstStep(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, fastScreen):
  if int(fastScreen):
    controller.show_frame("HomegeneousWellsLayout")
  else:
    self.configFile["extractAdvanceZebraParameters"] = 0
    if int(freeZebra):
      controller.show_frame("FreelySwimmingExperiment")
    else:
      chooseGeneralExperiment(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, 0)

def chooseGeneralExperiment(self, controller, freeZebra, headEmbZebra, drosophilia, rodent, other, freeZebra2):
  self.configFile["extractAdvanceZebraParameters"] = 0
  if int(freeZebra):
    self.organism = 'zebrafish'
    self.configFile["headEmbeded"] = 0
    controller.show_frame("WellOrganisation")
  elif int(freeZebra2):
    self.organism = 'zebrafishNew'
    self.configFile["headEmbeded"] = 0
    controller.show_frame("WellOrganisation")
  elif int(headEmbZebra):
    self.organism = 'headembeddedzebrafish'
    self.configFile["headEmbeded"] = 1
    chooseBeginningAndEndOfVideo(self, controller)
  else:
    self.organism = 'drosoorrodent'
    self.configFile["headEmbeded"] = 0
    self.configFile["freeSwimmingTailTrackingMethod"] = "none"
    controller.show_frame("WellOrganisation")

def wellOrganisation(self, controller, circular, rectangular, roi, other, multipleROIs, groupSameSizeAndShapeEquallySpacedWells):
  if multipleROIs:
    controller.show_frame("NbRegionsOfInterest")
  else:
    if groupSameSizeAndShapeEquallySpacedWells:
      self.shape = 'groupSameSizeAndShapeEquallySpacedWells'
      self.configFile["groupOfMultipleSameSizeAndShapeEquallySpacedWells"] = 1
      controller.show_frame("CircularOrRectangularWells")
    else:
      if rectangular:
        self.shape = 'rectangular'
        self.configFile["wellsAreRectangles"] = 1
        controller.show_frame("CircularOrRectangularWells")
      else:
        if circular and self.organism != 'drosoorrodent': # should remove the self.organism != 'drosoorrodent' at some point
          self.shape = 'circular'
          controller.show_frame("CircularOrRectangularWells")
        else:
          self.shape = 'other'
          if roi:
            cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
            cap.set(1, 10)
            ret, frame = cap.read()
            frame2 = frame.copy()
            [frame2, getRealValueCoefX, getRealValueCoefY, horizontal, vertical] = resizeImageTooLarge(frame2, True, 0.85)
            
            WINDOW_NAME = "Click on the top left of the region of interest"
            cvui.init(WINDOW_NAME)
            cv2.moveWindow(WINDOW_NAME, 0,0)
            cvui.imshow(WINDOW_NAME, frame2)
            while not(cvui.mouse(cvui.CLICK)):
              cursor = cvui.mouse()
              frame3 = frame2.copy()
              frame3 = cv2.line(frame3, (cursor.x - 2000, cursor.y), (cursor.x + 2000, cursor.y), (0, 0, 255), 2)
              frame3 = cv2.line(frame3, (cursor.x, cursor.y - 2000), (cursor.x, cursor.y + 2000), (0, 0, 255), 2)
              cvui.imshow(WINDOW_NAME, frame3)
              if cv2.waitKey(20) == 27:
                break
            self.configFile["oneWellManuallyChosenTopLeft"] = [int(getRealValueCoefX * cursor.x), int(getRealValueCoefY * cursor.y)]
            cv2.destroyAllWindows()
            
            WINDOW_NAME = "Click on the bottom right of the region of interest"
            cvui.init(WINDOW_NAME)
            cv2.moveWindow(WINDOW_NAME, 0,0)
            cvui.imshow(WINDOW_NAME, frame2)
            while not(cvui.mouse(cvui.CLICK)):
              cursor = cvui.mouse()
              frame3 = frame2.copy()
              frame3 = cv2.line(frame3, (cursor.x - 2000, cursor.y), (cursor.x + 2000, cursor.y), (0, 0, 255), 2)
              frame3 = cv2.line(frame3, (cursor.x, cursor.y - 2000), (cursor.x, cursor.y + 2000), (0, 0, 255), 2)
              cvui.imshow(WINDOW_NAME, frame3)
              if cv2.waitKey(20) == 27:
                break
            self.configFile["oneWellManuallyChosenBottomRight"] = [int(getRealValueCoefX * cursor.x), int(getRealValueCoefY * cursor.y)]
            cv2.destroyAllWindows()
            
            self.configFile["nbWells"] = 1
            chooseBeginningAndEndOfVideo(self, controller)
          else:
            self.configFile["noWellDetection"] = 1
            self.configFile["nbWells"] = 1
            chooseBeginningAndEndOfVideo(self, controller)


def regionsOfInterest(self, controller, nbwells):
  
  self.configFile["multipleROIsDefinedDuringExecution"] = 1
  self.configFile["nbWells"] = int(nbwells)
  
  chooseBeginningAndEndOfVideo(self, controller)
  

def rectangularWells(self, controller, nbwells, nbRowsOfWells, nbWellsPerRows):
  
  [pathToVideo, videoName, videoExt, configFile, argv] = getMainArguments(self)
  
  configFile["adjustRectangularWellsDetect"] = 1
  
  try:
    WINDOW_NAME = "Please Wait"
    cvui.init(WINDOW_NAME)
    cv2.moveWindow(WINDOW_NAME, 0,0)
    r = cv2.waitKey(2)
    mainZZ(pathToVideo, videoName, videoExt, configFile, argv)
    cv2.destroyAllWindows()
  except ValueError:
    newhyperparameters = pickle.load(open('newhyperparameters', 'rb'))
    for index in newhyperparameters:
      configFile[index] = newhyperparameters[index]
  
  configFile["adjustRectangularWellsDetect"] = 0
  
  self.configFile = configFile
  
  chooseBeginningAndEndOfVideo(self, controller)


def homegeneousWellsLayout(self, controller, nbwells, nbRowsOfWells, nbWellsPerRows):

  self.configFile = {"nbAnimalsPerWell": 1, "nbWells": 8, "nbRowsOfWells": 2, "nbWellsPerRows": 4, "groupOfMultipleSameSizeAndShapeEquallySpacedWells": 1, "postProcessMultipleTrajectories": 1,   "postProcessRemoveLowProbabilityDetection" : 1, "postProcessLowProbabilityDetectionPercentOfMaximum" : 0.2, "trackingPointSizeDisplay": 4, "extractAdvanceZebraParameters": 0,  "validationVideoPlotHeading": 0, "trackTail": 0, "freqAlgoPosFollow": 100, "fasterMultiprocessing": 1, "copyOriginalVideoToOutputFolderForValidation": 0, "backgroundSubtractorKNN": 1}
  
  self.configFile["nbWells"]          = int(nbwells)
  
  if len(nbRowsOfWells):
    self.configFile["nbRowsOfWells"]  = int(nbRowsOfWells)
  else:
    self.configFile["nbRowsOfWells"]  = 1
    
  if len(nbWellsPerRows):
    self.configFile["nbWellsPerRows"]  = int(nbWellsPerRows)
  else:
    self.configFile["nbWellsPerRows"]  = 4
  
  controller.show_frame("FinishConfig")


def morePreciseFastScreen(self, controller, nbwells, nbRowsOfWells, nbWellsPerRows):

  # The gaussian image filtering should be added here in the future
  self.configFile = {"minPixelDiffForBackExtract": 20, "backgroundPreProcessParameters": [[3]], "backgroundPreProcessMethod": ["erodeThenMin"], "trackingPointSizeDisplay": 1, "nbAnimalsPerWell": 1, "extractAdvanceZebraParameters": 0, "trackTail": 0, "nbWells": 1, "noWellDetection": 1, "backgroundExtractionForceUseAllVideoFrames": 1, "headSize": 2, "createValidationVideo": 0, "lastFrame": 1000}
  
  self.configFile["nbAnimalsPerWell"] = int(nbwells)
  
  # self.configFile["nbWells"]          = int(nbwells)
  
  if len(nbRowsOfWells):
    self.configFile["nbRowsOfWells"]  = int(nbRowsOfWells)
  else:
    self.configFile["nbRowsOfWells"]  = 1
    
  if len(nbWellsPerRows):
    self.configFile["nbWellsPerRows"]  = int(nbWellsPerRows)
  else:
    self.configFile["nbWellsPerRows"]  = 4
  
  self.calculateBackgroundFreelySwim(controller, 0, True)
  
  # controller.show_frame("FinishConfig")


def circularOrRectangularWells(self, controller, nbwells, nbRowsOfWells, nbWellsPerRows):
  self.configFile["nbWells"]        = int(nbwells)
  
  if len(nbRowsOfWells):
    self.configFile["nbRowsOfWells"]  = int(nbRowsOfWells)
  else:
    self.configFile["nbRowsOfWells"]  = 1
    
  if len(nbWellsPerRows):
    self.configFile["nbWellsPerRows"]  = int(nbWellsPerRows)
  else:
    self.configFile["nbWellsPerRows"]  = 4
  
  if self.shape == 'circular':
    controller.show_frame("ChooseCircularWellsLeft")
  else:
    if self.shape == 'groupSameSizeAndShapeEquallySpacedWells':
      chooseBeginningAndEndOfVideo(self, controller)
    else:
      rectangularWells(self, controller, nbwells, nbRowsOfWells, nbWellsPerRows)


def chooseCircularWellsLeft(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  ret, frame = cap.read()
  [x, y] = findWellLeft(frame)
  self.wellLeftBorderX = x
  self.wellLeftBorderY = y
  controller.show_frame("ChooseCircularWellsRight")
  
def chooseCircularWellsRight(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  ret, frame = cap.read()
  [xRight, yRight] = findWellRight(frame)
  xLeft = self.wellLeftBorderX
  yLeft = self.wellLeftBorderY
  dist = math.sqrt((xLeft - xRight)**2 + (yLeft - yRight)**2)
  self.configFile["minWellDistanceForWellDetection"] = int(dist)
  self.configFile["wellOutputVideoDiameter"]         = int(dist + dist * 0.2)
  chooseBeginningAndEndOfVideo(self, controller)

def chooseBeginningAndEndOfVideo(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  max_l = int(cap.get(7)) - 2
  
  cap.set(1, 1)
  ret, frame = cap.read()
  WINDOW_NAME = "Choose where the analysis of your video should start."
  WINDOW_NAME_CTRL = "Control"
  cvui.init(WINDOW_NAME)
  cv2.moveWindow(WINDOW_NAME, 0,0)
  cvui.init(WINDOW_NAME_CTRL)
  cv2.moveWindow(WINDOW_NAME_CTRL, 0, 300)
  value = [1]
  curValue = value[0]
  buttonclicked = False
  buttonEntireVideo = False
  widgetX = 40
  widgetY = 20
  widgetL = 300
  while not(buttonclicked) and not(buttonEntireVideo):
      value[0] = int(value[0])
      if curValue != value[0]:
        cap.set(1, value[0])
        frameOld = frame
        ret, frame = cap.read()
        if not(ret):
          frame = frameOld
        curValue = value[0]
      frameCtrl = np.full((200, 750), 100).astype('uint8')
      frameCtrl[widgetY:widgetY+60, widgetX:widgetX+widgetL] = 0
      cvui.text(frameCtrl, widgetX, widgetY, 'Frame')
      cvui.trackbar(frameCtrl, widgetX, widgetY+10, widgetL, value, 0, max_l)
      cvui.counter(frameCtrl, widgetX, widgetY+60, value)
      buttonclicked = cvui.button(frameCtrl, widgetX, widgetY+90, "Ok, I want the tracking to start at this frame!")
      
      cvui.text(frameCtrl, widgetX+350, widgetY+1, 'No, this is unecessary:')
      buttonEntireVideo = cvui.button(frameCtrl, widgetX+350, widgetY+40, "I want the tracking to run on the entire video!")
      
      cvui.text(frameCtrl, widgetX, widgetY+130, 'Keys: 4 or a: move backwards; 6 or d: move forward')
      cvui.text(frameCtrl, widgetX, widgetY+160, 'Keys: g or f: fast backwards; h or j: fast forward')
      frame2 = frame.copy()
      [frame2, getRealValueCoefX, getRealValueCoefY, horizontal, vertical] = resizeImageTooLarge(frame2, True, 0.85)
      cvui.imshow(WINDOW_NAME, frame2)
      cvui.imshow(WINDOW_NAME_CTRL, frameCtrl)
      r = cv2.waitKey(20)
      if (r == 54) or (r == 100) or (r == 0):
        value[0] = value[0] + 1
      elif (r == 52) or (r == 97) or (r == 113):
        value[0] = value[0] - 1
      elif (r == 103):
        value[0] = value[0] - 30
      elif (r == 104):
        value[0] = value[0] + 30
      elif (r == 102):
        value[0] = value[0] - 100
      elif (r == 106):
        value[0] = value[0] + 100
  cv2.destroyAllWindows()
  
  if not(buttonEntireVideo):
    self.configFile["firstFrame"] = int(value[0])
    cap.set(1, max_l)
    ret, frame = cap.read()
    while not(ret):
      max_l = max_l - 1
      cap.set(1, max_l)
      ret, frame = cap.read()
    WINDOW_NAME = "Choose where the analysis of your video should end."
    WINDOW_NAME_CTRL = "Control"
    cvui.init(WINDOW_NAME)
    cv2.moveWindow(WINDOW_NAME, 0,0)
    cvui.init(WINDOW_NAME_CTRL)
    cv2.moveWindow(WINDOW_NAME_CTRL, 0, 300)
    value = [max_l]
    curValue = value[0]
    buttonclicked = False
    widgetX = 40
    widgetY = 20
    widgetL = 300
    while not(buttonclicked):
        value[0] = int(value[0])
        if curValue != value[0]:
          cap.set(1, value[0])
          frameOld = frame
          ret, frame = cap.read()
          if not(ret):
            frame = frameOld
          curValue = value[0]
        frameCtrl = np.full((200, 400), 100).astype('uint8')
        frameCtrl[widgetY:widgetY+60, widgetX:widgetX+widgetL] = 0
        cvui.text(frameCtrl, widgetX, widgetY, 'Frame')
        cvui.trackbar(frameCtrl, widgetX, widgetY+10, widgetL, value, 0, max_l-1)
        cvui.counter(frameCtrl, widgetX, widgetY+60, value)
        buttonclicked = cvui.button(frameCtrl, widgetX, widgetY+90, "Ok, I want the tracking to end at this frame!")
        cvui.text(frameCtrl, widgetX, widgetY+130, 'Keys: 4 or a: move backwards; 6 or d: move forward')
        cvui.text(frameCtrl, widgetX, widgetY+160, 'Keys: g or f: fast backwards; h or j: fast forward')
        frame2 = frame.copy()
        [frame2, getRealValueCoefX, getRealValueCoefY, horizontal, vertical] = resizeImageTooLarge(frame2, True, 0.85)
        cvui.imshow(WINDOW_NAME, frame2)
        cvui.imshow(WINDOW_NAME_CTRL, frameCtrl)
        r = cv2.waitKey(20)
        if (r == 54) or (r == 100) or (r == 0):
          value[0] = value[0] + 1
        elif (r == 52) or (r == 97) or (r == 113):
          value[0] = value[0] - 1
        elif (r == 103):
          value[0] = value[0] - 30
        elif (r == 104):
          value[0] = value[0] + 30
        elif (r == 102):
          value[0] = value[0] - 100
        elif (r == 106):
          value[0] = value[0] + 100
    self.configFile["lastFrame"] = int(value[0])
    cv2.destroyAllWindows()
  
  if int(self.configFile["headEmbeded"]) == 1:
    controller.show_frame("HeadEmbeded")
  else:
    if self.organism == 'zebrafishNew':
      controller.show_frame("NumberOfAnimals2")
    elif self.organism == 'zebrafish':
      controller.show_frame("NumberOfAnimals")
    else:
      controller.show_frame("NumberOfAnimalsCenterOfMass")

  
def getImageForMultipleAnimalGUI(l, vertical, horizontal, nx, ny, max_l, videoToCreateConfigFileFor, background, wellPositions, hyperparameters):
  
  [frame, a1, a2] = getForegroundImage(videoToCreateConfigFileFor, background, l, 0, [], hyperparameters)
  
  lengthX = nx * 2
  lengthY = ny
  
  newX = lengthX
  newY = lengthY
  
  vertical2   = vertical   - vertical   * 0.12
  horizontal2 = horizontal - horizontal * 0.01
  if ( (lengthX > horizontal2) or (lengthY > vertical2) ):
    sinkFactor = 1
    sinkFactorX = horizontal2 / lengthX
    sinkFactorY = vertical2   / lengthY
    if (sinkFactorX > sinkFactorY):
      sinkFactor = sinkFactorY
    else:
      sinkFactor = sinkFactorX
    newX = lengthX * sinkFactor
    newY = lengthY * sinkFactor
  
  frame2 = frame
  ret,thresh2 = cv2.threshold(frame2,hyperparameters["thresholdForBlobImg"],255,cv2.THRESH_BINARY)
  kernel  = np.ones((hyperparameters["erodeSize"],hyperparameters["erodeSize"]), np.uint8)
  thresh2 = cv2.dilate(thresh2, kernel, iterations=hyperparameters["dilateIter"])

  thresh = thresh2
  thresh2 = cv2.cvtColor(thresh2, cv2.COLOR_GRAY2RGB)
  frame   = cv2.cvtColor(frame,   cv2.COLOR_GRAY2RGB)
  thresh[0, :]                = 255
  thresh[:, 0]                = 255
  thresh[len(thresh)-1, :]    = 255
  thresh[:, len(thresh[0])-1] = 255
  areaList = []
  contours, hierarchy = cv2.findContours(thresh,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
  for contour in contours:
    area = cv2.contourArea(contour)
    if area < (len(thresh) * len(thresh[0]))/2:
      areaList.append(area)
    if (area > hyperparameters["minArea"]) and (area < hyperparameters["maxArea"]):
      M = cv2.moments(contour)
      if M['m00']:
        x = int(M['m10']/M['m00'])
        y = int(M['m01']/M['m00'])
        cv2.circle(thresh2, (x,y), 3, (0,0,255), -1)
      else:
        x = 0
        y = 0
  
  frame = cv2.line(frame, (len(frame[0])-5, 0), (len(frame[0])-5, len(frame)), (255, 0, 0), 5) 
  
  frame = np.concatenate((frame, thresh2), axis=1)
  
  frame = cv2.resize(frame,(int(newX),int(newY)))
  
  if len(areaList):
    maxToReturn = int((max(areaList)+2)*2)
  else:
    maxToReturn = (len(thresh) * len(thresh[0]))/5
  
  return [frame, maxToReturn]

def printStuffOnCtrlImg(frameCtrl, frameNum, x, y, l, minn, maxx, name):
  cvui.text(frameCtrl,     x,         y,    name)
  cvui.trackbar(frameCtrl, x,      y+10, l, frameNum, minn, maxx)
  cvui.counter(frameCtrl,  x+l+10, y+20,    frameNum)

def identifyMultipleHead(self, controller, nbanimals):
  
  self.configFile["videoName"] = "configFilePrep"

  tempConfig = self.configFile
  
  horizontal = self.winfo_screenwidth()
  vertical   = self.winfo_screenheight()
  
  # Wait image
  WINDOW_NAME = "Please Wait"
  cv2.destroyAllWindows()
  cvui.init(WINDOW_NAME)
  cv2.moveWindow(WINDOW_NAME, 0,0)
  # Getting hyperparameters, wellPositions, and background
  hyperparameters = getHyperparametersSimple(tempConfig)
  wellPositions = findWells(self.videoToCreateConfigFileFor, hyperparameters)
  background    = getBackground(self.videoToCreateConfigFileFor, hyperparameters)
  
  cur_dir_path = os.path.dirname(os.path.realpath(__file__))
  cur_dir_path = Path(cur_dir_path)
  
  tab = [1]
  img = cv2.imread(os.path.join(cur_dir_path, 'no1.png'))
  img = cv2.resize(img,(int(horizontal*0.95),int(vertical*0.8)))
  buttonclicked = False
  count = 0
  while not(buttonclicked):
    buttonclicked = cvui.button(img, 10, 10, "Ok, I understand!")
    cvui.imshow(WINDOW_NAME, img)
    cv2.waitKey(20)
    count = count + 1
    if count > 100:
      buttonclicked = True
  img = cv2.imread(os.path.join(cur_dir_path, 'no2.png'))
  img = cv2.resize(img,(int(horizontal*0.95),int(vertical*0.8)))
  buttonclicked = False
  count = 0
  while not(buttonclicked):
    buttonclicked = cvui.button(img, 10, 10, "Ok, I understand!")
    cvui.imshow(WINDOW_NAME, img)
    cv2.waitKey(20)
    count = count + 1
    if count > 100:
      buttonclicked = True
  img = cv2.imread(os.path.join(cur_dir_path, 'ok1.png'))
  img = cv2.resize(img,(int(horizontal*0.95),int(vertical*0.8)))
  buttonclicked = False
  count = 0
  while not(buttonclicked):
    buttonclicked = cvui.button(img, 10, 10, "Ok, I understand!")
    cvui.imshow(WINDOW_NAME, img)
    cv2.waitKey(20)
    count = count + 1
    if count > 100:
      buttonclicked = True
  
  WINDOW_NAME = "Adjust Parameters: As much as possible, you must see red points on and only on animals on the right image."
  WINDOW_NAME_CTRL = "Adjust Parameters."
  cv2.destroyAllWindows()
  # Manual parameters adjustements
  cap        = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  nx         = int(cap.get(3))
  ny         = int(cap.get(4))
  max_l      = int(cap.get(7))
  
  hyperparameters["minArea"] = 5
  hyperparameters["maxArea"] = 800

  [frame, maxAreaBlobs] = getImageForMultipleAnimalGUI(1, vertical, horizontal, nx, ny, max_l, self.videoToCreateConfigFileFor, background, wellPositions, hyperparameters)
  frameCtrl = np.full((200, 1100), 100).astype('uint8')
  
  cvui.init(WINDOW_NAME)
  cv2.moveWindow(WINDOW_NAME, 0, 0)
  cvui.imshow(WINDOW_NAME, frame)
  
  cvui.init(WINDOW_NAME_CTRL)
  cv2.moveWindow(WINDOW_NAME_CTRL, 0, vertical-290)
  cvui.imshow(WINDOW_NAME_CTRL, frameCtrl)
  
  frameNum = [hyperparameters["firstFrame"]] if "firstFrame" in hyperparameters else [ 1 ]
  curFrameNum = frameNum[0] + 1
  minPixelDiffForBackExtract = [hyperparameters["minPixelDiffForBackExtract"]]
  thresholdForBlobImg        = [hyperparameters["thresholdForBlobImg"]]
  dilateIter                 = [hyperparameters["dilateIter"]]
  minArea                    = [hyperparameters["minArea"]]
  maxArea                    = [hyperparameters["maxArea"]]
  firstFrame = hyperparameters["firstFrame"] if "firstFrame" in hyperparameters else 1
  lastFrame  = hyperparameters["lastFrame"]-1 if "lastFrame" in hyperparameters else max_l - 10
  
  buttonclicked = False
  while not(buttonclicked):
    if curFrameNum != frameNum[0] or hyperparameters["minPixelDiffForBackExtract"] != minPixelDiffForBackExtract[0] or hyperparameters["thresholdForBlobImg"] != thresholdForBlobImg[0] or hyperparameters["dilateIter"] != dilateIter[0] or hyperparameters["minArea"] != minArea[0] or hyperparameters["maxArea"] != maxArea[0]:
      
      curFrameNum = frameNum[0]
      hyperparameters["minPixelDiffForBackExtract"] = int(minPixelDiffForBackExtract[0])
      hyperparameters["thresholdForBlobImg"] = int(thresholdForBlobImg[0])
      hyperparameters["dilateIter"] = int(dilateIter[0])
      hyperparameters["minArea"] = int(minArea[0])
      hyperparameters["maxArea"] = int(maxArea[0])
      
      [frame, maxAreaBlobs] = getImageForMultipleAnimalGUI(curFrameNum, vertical, horizontal, nx, ny, max_l, self.videoToCreateConfigFileFor, background, wellPositions, hyperparameters)
      
    frameCtrl = np.full((200, 1100), 100).astype('uint8')
    
    printStuffOnCtrlImg(frameCtrl, frameNum,                     1, 5,  350,  firstFrame, lastFrame, "Frame number")
    printStuffOnCtrlImg(frameCtrl, minPixelDiffForBackExtract, 470, 5,  350,  0, 255, "Threshold left image")
    printStuffOnCtrlImg(frameCtrl, thresholdForBlobImg,          1, 71,  350, 0, 255, "Threshold right image")
    printStuffOnCtrlImg(frameCtrl, dilateIter,                 470, 71,  350, 0, 15, "Area dilatation")
    printStuffOnCtrlImg(frameCtrl, minArea,                      1, 137, 350, 0, maxAreaBlobs, "Minimum area")
    printStuffOnCtrlImg(frameCtrl, maxArea,                    470, 137, 350, 0, maxAreaBlobs, "Maximum area")
    
    buttonclicked = cvui.button(frameCtrl, 940, 10, "Ok, done!")
    
    cvui.imshow(WINDOW_NAME, frame)
    cvui.imshow(WINDOW_NAME_CTRL, frameCtrl)
    
    if cv2.waitKey(20) == 27:
        break
  cv2.destroyAllWindows()
  
  del self.configFile["videoName"]
  
  self.configFile["minPixelDiffForBackExtract"] = int(hyperparameters["minPixelDiffForBackExtract"])
  self.configFile["thresholdForBlobImg"]        = int(hyperparameters["thresholdForBlobImg"])
  self.configFile["dilateIter"]                 = int(hyperparameters["dilateIter"])
  self.configFile["minArea"]                    = int(hyperparameters["minArea"])
  self.configFile["maxArea"]                    = int(hyperparameters["maxArea"])
  self.configFile["headSize"]        = math.sqrt((int(hyperparameters["minArea"]) + int(hyperparameters["maxArea"])) / 2)


def numberOfAnimals(self, controller, nbanimals, yes, noo, forceBlobMethodForHeadTracking, yesBouts, nooBouts, recommendedMethod, alternativeMethod, yesBends, nooBends, adjustBackgroundExtractionBasedOnNumberOfBlackPixels):

  self.configFile["noBoutsDetection"] = 1
  self.configFile["noChecksForBoutSelectionInExtractParams"] = 1
  self.configFile["trackingPointSizeDisplay"] = 4
  self.configFile["validationVideoPlotHeading"] = 0
  
  if int(yesBends):
    self.configFile["extractAdvanceZebraParameters"] = 1
  
  nbanimals = int(nbanimals)
  if nbanimals == self.configFile["nbWells"]:
    self.configFile["nbAnimalsPerWell"] = 1
  else:
    self.configFile["nbAnimalsPerWell"] = int(nbanimals / self.configFile["nbWells"])
    if noo:
      self.configFile["multipleHeadTrackingIterativelyRelaxAreaCriteria"] = 0
    else:
      self.configFile["multipleHeadTrackingIterativelyRelaxAreaCriteria"] = 1
      
  self.forceBlobMethodForHeadTracking = int(forceBlobMethodForHeadTracking)
  if self.forceBlobMethodForHeadTracking:
    self.configFile["forceBlobMethodForHeadTracking"] = self.forceBlobMethodForHeadTracking
  
  if self.organism == 'zebrafish':
    controller.show_frame("IdentifyHeadCenter")
  elif self.organism == 'zebrafishNew':
    detectBouts = 0
    if int(yesBouts):
      detectBouts = 1
    method = 0
    if int(alternativeMethod):
      method = 1
    automaticallyFindOptimalParameters(self, controller, True, detectBouts, method, int(noo), int(adjustBackgroundExtractionBasedOnNumberOfBlackPixels))
  elif self.organism == 'drosoorrodent' and int(recommendedMethod):
    automaticallyFindOptimalParameters(self, controller, True, 0, 0, int(noo), int(adjustBackgroundExtractionBasedOnNumberOfBlackPixels))
  else:
    identifyMultipleHead(self, controller, nbanimals)
    controller.show_frame("FinishConfig")

def chooseHeadCenter(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  ret, frame = cap.read()
  [x, y] = findHeadCenter(frame)
  self.headCenterX = x
  self.headCenterY = y
  controller.show_frame("IdentifyBodyExtremity")

def chooseBodyExtremity(self, controller):
  cap = zzVideoReading.VideoCapture(self.videoToCreateConfigFileFor)
  ret, frame = cap.read()
  [extX, extY] = findBodyExtremity(frame)
  headCenterX = self.headCenterX
  headCenterY = self.headCenterY
  dist = math.sqrt((extX - headCenterX)**2 + (extY - headCenterY)**2)
  
  if self.organism == 'zebrafish':
    minArea = int(dist * (dist * 0.1))
    maxArea = int(dist * (dist * 0.4))
    self.configFile["minArea"]     = minArea
    self.configFile["maxArea"]     = maxArea
    self.configFile["minAreaBody"] = minArea
    self.configFile["maxAreaBody"] = maxArea
    self.configFile["headSize"]    = int(dist * 0.3)
    self.configFile["minTailSize"] = int(dist * 0.05)
    self.configFile["maxTailSize"] = int(dist * 2)
    self.configFile["paramGaussianBlur"] = int(int(dist*(31/52))/2)*2 + 1
  else:
    minArea = int(((2 * dist) * (2 * dist)) * 0.2)
    maxArea = int(((2 * dist) * (2 * dist)) * 1.5)
    self.configFile["minArea"]     = minArea
    self.configFile["maxArea"]     = maxArea
    self.configFile["minAreaBody"] = minArea
    self.configFile["maxAreaBody"] = maxArea
    self.configFile["headSize"]    = int(dist * 2)
  
  self.configFile["extractBackWhiteBackground"] = 1
  
  self.configFile["noBoutsDetection"] = 1
  
  if int(self.configFile["nbAnimalsPerWell"]) > 1 or self.forceBlobMethodForHeadTracking:
    identifyMultipleHead(self, controller, int(self.configFile["nbAnimalsPerWell"]))
    
  controller.show_frame("GoToAdvanceSettings")


def goToAdvanceSettings(self, controller, yes, no):

  if int(no):
    controller.show_frame("FinishConfig")
  else:
    self.configFile["noBoutsDetection"] = 0
    self.calculateBackgroundFreelySwim(controller, 0)

def finishConfig(self, controller, configFileNameToSave):

  cur_dir_path = os.path.dirname(os.path.realpath(__file__))
  cur_dir_path = Path(cur_dir_path)
  cur_dir_path = cur_dir_path.parent.parent
  reference = os.path.join(cur_dir_path, os.path.join('configuration', configFileNameToSave + '.json'))

  # Ideally would like to remove these four lines below, once the problem with wrong 'firstFrame' and 'lastFrame' being saved in the configuration file is solved
  if "lastFrame" in self.configFile:
    del self.configFile["lastFrame"]
  if "firstFrame" in self.configFile:
    del self.configFile["firstFrame"]
  
  with open(reference, 'w') as outfile:
    json.dump(self.configFile, outfile)
  self.configFile = {}
  self.videoToCreateConfigFileFor = ''
  self.wellLeftBorderX = 0
  self.wellLeftBorderY = 0
  self.headCenterX = 0
  self.headCenterY = 0
  self.organism = ''
  
  if (globalVariables["mac"] or globalVariables["lin"]):
    self.destroy()
  else:
    controller.show_frame("StartPage")



from zebrazoom.code.preprocessImage import preprocessImage
import numpy as np
import cv2

def getForegroundImageSequential(cap, videoPath, background, frameNumber, wellNumber, wellPositions, hyperparameters, alreadyExtractedImage=0, trackingHeadTailAllAnimals=0):
  
  minPixelDiffForBackExtract = hyperparameters["minPixelDiffForBackExtract"]
  if "minPixelDiffForBackExtractHead" in hyperparameters:
    minPixelDiffForBackExtract = hyperparameters["minPixelDiffForBackExtractHead"]
  debug = 0
  
  xtop = wellPositions[wellNumber]['topLeftX']
  ytop = wellPositions[wellNumber]['topLeftY']
  lenX = wellPositions[wellNumber]['lengthX']
  lenY = wellPositions[wellNumber]['lengthY']
  
  back = background[ytop:ytop+lenY, xtop:xtop+lenX]
  
  if (type(alreadyExtractedImage) != int):
    ret   = True
    frame = alreadyExtractedImage.copy()
  else:
    ret, frame = cap.read()
  
  if not(ret):
    currentFrameNum = int(cap.get(1))
    while not(ret):
      currentFrameNum = currentFrameNum - 1
      cap.set(1, currentFrameNum)
      ret, frame = cap.read()
      
  if hyperparameters["invertBlackWhiteOnImages"]:
    frame = 255 - frame
  
  if hyperparameters["imagePreProcessMethod"]:
    frame = preprocessImage(frame, hyperparameters)
  
  grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
  curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
  
  if hyperparameters["trackOnlyOnROI_halfDiameter"] != 0 and frameNumber != hyperparameters["firstFrame"]:
    xHead = trackingHeadTailAllAnimals[0][frameNumber - hyperparameters["firstFrame"] - 1][0][0]
    yHead = trackingHeadTailAllAnimals[0][frameNumber - hyperparameters["firstFrame"] - 1][0][1]
    xHeadMin = xHead
    xHeadMax = xHead
    yHeadMin = yHead
    yHeadMax = yHead
    for animalId in range(1, hyperparameters["nbAnimalsPerWell"]):
      xHead = trackingHeadTailAllAnimals[animalId][frameNumber - hyperparameters["firstFrame"] - 1][0][0]
      yHead = trackingHeadTailAllAnimals[animalId][frameNumber - hyperparameters["firstFrame"] - 1][0][1]
      xHeadMin = xHeadMin if xHeadMin < xHead else xHead
      xHeadMax = xHeadMax if xHeadMax > xHead else xHead
      yHeadMin = yHeadMin if yHeadMin < yHead else yHead
      yHeadMax = yHeadMax if yHeadMax > yHead else yHead
    maxHalfDiameter = hyperparameters["trackOnlyOnROI_halfDiameter"]
    xmin = int(xHeadMin - maxHalfDiameter) if xHeadMin - maxHalfDiameter >= 0 else 0
    ymin = int(yHeadMin - maxHalfDiameter) if yHeadMin - maxHalfDiameter >= 0 else 0
    xmax = int(xHeadMax + maxHalfDiameter) if xHeadMax + maxHalfDiameter < len(curFrame[0]) else len(curFrame[0]) - 1
    ymax = int(yHeadMax + maxHalfDiameter) if yHeadMax + maxHalfDiameter < len(curFrame)    else len(curFrame) - 1
    curFrameInitial = curFrame.copy()
    backInitial     = back.copy()
    initialCurFrameInitial = curFrame.copy()
    curFrame = curFrame[ymin:ymax, xmin:xmax]
    back     = back[ymin:ymax, xmin:xmax]
  else:
    xmin = 0
    ymin = 0
    initialCurFrameInitial = 0
    curFrameInitial = 0
    backInitial     = 0
  
  initialCurFrame = curFrame.copy()
  
  putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
  
  curFrame[putToWhite] = 255
  previousNbBlackPixels = []
  if hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"]:
    minPixel2nbBlackPixels = {}
    countTries = 0
    nbBlackPixels = 0
    nbBlackPixelsMax = int(hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"])
    while (minPixelDiffForBackExtract > 0) and (countTries < 30) and not(minPixelDiffForBackExtract in minPixel2nbBlackPixels):
      if countTries > 0:
        curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
        if hyperparameters["trackOnlyOnROI_halfDiameter"] != 0 and frameNumber != hyperparameters["firstFrame"]:
          curFrame = curFrame[ymin:ymax, xmin:xmax]
        putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
        curFrame[putToWhite] = 255
      ret, thresh1 = cv2.threshold(curFrame, hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      thresh1 = 255 - thresh1
      nbBlackPixels = cv2.countNonZero(thresh1)
      minPixel2nbBlackPixels[minPixelDiffForBackExtract] = nbBlackPixels
      if nbBlackPixels > nbBlackPixelsMax:
        minPixelDiffForBackExtract = minPixelDiffForBackExtract + 1
      if nbBlackPixels <= nbBlackPixelsMax:
        minPixelDiffForBackExtract = minPixelDiffForBackExtract - 1
      countTries = countTries + 1
      previousNbBlackPixels.append(nbBlackPixels)
      if len(previousNbBlackPixels) >= 3:
        lastThree = previousNbBlackPixels[len(previousNbBlackPixels)-3: len(previousNbBlackPixels)]
        if lastThree.count(lastThree[0]) == len(lastThree):
          countTries = 1000000
    
    best_minPixelDiffForBackExtract = 0
    minDist = 10000000000000
    for minPixelDiffForBackExtract in minPixel2nbBlackPixels:
      nbBlackPixels = minPixel2nbBlackPixels[minPixelDiffForBackExtract]
      dist = abs(nbBlackPixels - nbBlackPixelsMax)
      if dist < minDist:
        minDist = dist
        best_minPixelDiffForBackExtract = minPixelDiffForBackExtract
        
    minPixelDiffForBackExtract = best_minPixelDiffForBackExtract
    hyperparameters["minPixelDiffForBackExtractHead"] = minPixelDiffForBackExtract
    curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX]
    if hyperparameters["trackOnlyOnROI_halfDiameter"] != 0 and frameNumber != hyperparameters["firstFrame"]:
      curFrame = curFrame[ymin:ymax, xmin:xmax]
      curFrameInitial = curFrame
      backInitial     = back
      initialCurFrameInitial = curFrame
    putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
    curFrame[putToWhite] = 255      
    
  # else:
    # putToBlack3a = ( curFrame.astype('int32') <= (back.astype('int32') - minPixelDiffForBackExtract) )
    # putToBlack3b = ( curFrame.astype('int32') >= (back.astype('int32') + minPixelDiffForBackExtract) )
    
    # putToWhite  = (curFrame <= 220)
    # putToBlack  = (curFrame >= 220)
    
    # curFrame[putToWhite]   = 255
    # curFrame[putToBlack]   = 0
    # curFrame[putToBlack3a] = 0
    # curFrame[putToBlack3b] = 0
  
  if (debug):
    # cv2.imshow('Frame', curFrame)
    cv2.imshow('Frame', frame)
    cv2.waitKey(0)
  
  if hyperparameters["trackOnlyOnROI_halfDiameter"] != 0:
    ret, res = cv2.threshold(curFrame, hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
    if ret:
      reinitialize = 1
      contours, hierarchy = cv2.findContours(res, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
      for contour in contours:
        contourArea = cv2.contourArea(contour)
        if contourArea > hyperparameters["minAreaBody"] and contourArea < hyperparameters["maxAreaBody"]:
          reinitialize = 0
      if reinitialize and type(curFrameInitial) != int:
        curFrame = curFrameInitial
        initialCurFrame = initialCurFrameInitial
        back = backInitial
  
  return [curFrame, initialCurFrame, back, xmin, ymin]



import os
import math
import shutil
import matplotlib.pyplot as plt
import pickle
import numpy as np
import pandas as pd
import json

def populationComparaison(nameOfFile, resFolder, globParam, conditions, genotypes, outputFolder, medianPerWellFirstForEachKinematicParameter = 0, plotOutliersAndMean = True, saveDataPlottedInJson = 0, medianPerGenotypeFirstForEachKinematicParameter=0):

  outputFolderResult = os.path.join(outputFolder, nameOfFile)
  if not(os.path.exists(outputFolderResult)):
    os.mkdir(outputFolderResult)
    
  if medianPerWellFirstForEachKinematicParameter:
    outputFolderResult = os.path.join(outputFolderResult, 'medianPerWellFirst')
  else:
    if medianPerGenotypeFirstForEachKinematicParameter:
      outputFolderResult = os.path.join(outputFolderResult, 'medianPerGenotypeFirst')
    else:
      outputFolderResult = os.path.join(outputFolderResult, 'allBoutsMixed')
  
  if plotOutliersAndMean: # This is a little bit of a hack because the creation of folders relies on the order in which the populationComparaison function is called (relative to the plotOutliersAndMean parameter)
    if os.path.exists(outputFolderResult): 
      shutil.rmtree(outputFolderResult)
    while True:
      try:
        os.mkdir(outputFolderResult)
        break
      except:
        print("waiting to create folder:", outputFolderResult)
  else:
    outputFolderResult = os.path.join(outputFolderResult, 'noMeanAndOutliersPlotted')
    os.mkdir(outputFolderResult)
  
  dataPlotted = {}
  
  if os.path.exists(os.path.join(resFolder, nameOfFile + '.pkl')):
    infile = open(os.path.join(resFolder, nameOfFile + '.pkl'),'rb')
  else:
    infile = open(os.path.join(resFolder, nameOfFile),'rb') # This is just to insure compatibility with previous versions, should remove this line in the future
  dfParam = pickle.load(infile)
  infile.close()
  
  columnsForRawDataExport = ['Trial_ID', 'Well_ID', 'NumBout', 'BoutStart', 'BoutEnd', 'Condition', 'Genotype'] + globParam
  
  if medianPerWellFirstForEachKinematicParameter:
    dfKinematicValues = dfParam[columnsForRawDataExport]
    dfKinematicValues = dfKinematicValues.astype({param: float for param in globParam})
    dfKinematicValues = dfKinematicValues.groupby(['Trial_ID', 'Well_ID']).median()
    dfCondGeno = dfParam[['Trial_ID', 'Well_ID', 'Condition', 'Genotype']]
    dfCondGeno = dfCondGeno.groupby(['Trial_ID', 'Well_ID']).first()
    dfCount = dfParam[['Trial_ID', 'Well_ID']].copy()
    dfCount['numberOfBouts_NoBoutsRemovedBasedOnBends'] = [0 for i in range(len(dfCount['Trial_ID']))]
    dfCount = dfCount.groupby(['Trial_ID', 'Well_ID']).count()
    dfParam = pd.concat([dfCondGeno, dfKinematicValues], axis=1)
    dfParam = pd.concat([dfParam, dfCount], axis=1)
    globParam = globParam + ['numberOfBouts_NoBoutsRemovedBasedOnBends']
  elif medianPerGenotypeFirstForEachKinematicParameter:
    dfKinematicValues = dfParam[columnsForRawDataExport]
    dfKinematicValues = dfKinematicValues.astype({param: float for param in globParam})
    dfKinematicValues = dfKinematicValues.groupby(['Genotype', 'Well_ID']).median()
    dfCond = dfParam[['Genotype', 'Well_ID', 'Condition']]
    dfCond = dfCond.groupby(['Genotype', 'Well_ID']).first()
    dfCount = dfParam[['Genotype', 'Well_ID']].copy()
    dfCount['numberOfBouts_NoBoutsRemovedBasedOnBends'] = [0 for i in range(len(dfCount['Genotype']))]
    dfCount = dfCount.groupby(['Genotype', 'Well_ID']).count()
    dfParam = pd.concat([dfCond, dfKinematicValues], axis=1)
    dfParam = pd.concat([dfParam, dfCount], axis=1)
    globParam = globParam + ['numberOfBouts_NoBoutsRemovedBasedOnBends']
  else:
    dfParam  = dfParam[columnsForRawDataExport]
  
  if plotOutliersAndMean:
    dfParam.to_excel(os.path.join(outputFolderResult, 'globalParametersInsideCategories.xlsx'))
  
  nbGraphs = int(len(globParam)/6) if len(globParam) % 6 == 0 else int(len(globParam)/6) + 1
  color = ['b', 'r', 'c', 'm', 'y', 'k']
  for i in range(nbGraphs):
    globParamForPlot = [globParam[elem] for elem in range(6*i, min(6*(i+1), len(globParam)))]
    nbLines   = int(math.sqrt(len(globParamForPlot)))
    nbColumns = math.ceil(len(globParamForPlot) / nbLines)
    fig, tabAx = plt.subplots(nbLines, nbColumns, figsize=(22.9, 8.8))
    for idx, parameter in enumerate(globParamForPlot):
      print("plotting parameter:", parameter)
      concatenatedValues = []
      labels = []
      if not(medianPerGenotypeFirstForEachKinematicParameter):
        for condition in conditions:
          for genotype in genotypes:
            indicesCondition = dfParam.index[dfParam['Condition'] == condition].tolist()
            indicesGenotype  = dfParam.index[dfParam['Genotype']  == genotype].tolist()
            indices = [ind for ind in indicesCondition if ind in indicesGenotype]
            values  = dfParam.loc[indices, parameter].values
            concatenatedValues.append(values)
            labels.append(str(condition) + '\n' + str(genotype))
      else:
        for condition in conditions:
          indicesCondition = dfParam.index[dfParam['Condition'] == condition].tolist()
          values  = dfParam.loc[indicesCondition, parameter].values
          concatenatedValues.append(values)
          labels.append(str(condition))
      
      concatenatedValuesWithoutNans = []
      for toConcat in concatenatedValues:
        concatenatedValuesWithoutNans.append(np.array([x for x in toConcat if not(math.isnan(x))]))
      concatenatedValues = concatenatedValuesWithoutNans
      
      if saveDataPlottedInJson:
        dataPlotted[parameter] = {}
        for idx2, label in enumerate(labels):
          dataPlotted[parameter][label] = concatenatedValues[idx2].tolist()
      
      if nbLines == 1:
        if nbColumns == 1:
          tabAx.set_title(parameter)
          tabAx.boxplot(concatenatedValues, showmeans=plotOutliersAndMean, showfliers=plotOutliersAndMean)
          if plotOutliersAndMean and (medianPerWellFirstForEachKinematicParameter or medianPerGenotypeFirstForEachKinematicParameter or len(concatenatedValues[0]) < 100):
            for idx2, values in enumerate(concatenatedValues):
              tabAx.plot(np.random.normal(idx2+1, 0.005*len(concatenatedValues), size=len(values)), values, 'b.', alpha=0.3, c=color[idx2] if idx2 < len(color) else 'b')
          tabAx.set_xticklabels(labels)
        else:
          tabAx[idx%nbColumns].set_title(parameter)
          tabAx[idx%nbColumns].boxplot(concatenatedValues, showmeans=plotOutliersAndMean, showfliers=plotOutliersAndMean)
          if plotOutliersAndMean and (medianPerWellFirstForEachKinematicParameter or medianPerGenotypeFirstForEachKinematicParameter or len(concatenatedValues[0]) < 100):
            for idx2, values in enumerate(concatenatedValues):
              tabAx[idx%nbColumns].plot(np.random.normal(idx2+1, 0.005*len(concatenatedValues), size=len(values)), values, 'b.', alpha=0.3, c=color[idx2] if idx2 < len(color) else 'b')
          tabAx[idx%nbColumns].set_xticklabels(labels)
      else:
        tabAx[int(idx/nbColumns), idx%nbColumns].set_title(parameter)
        tabAx[int(idx/nbColumns), idx%nbColumns].boxplot(concatenatedValues, showmeans=plotOutliersAndMean, showfliers=plotOutliersAndMean)
        if plotOutliersAndMean and (medianPerWellFirstForEachKinematicParameter or medianPerGenotypeFirstForEachKinematicParameter or len(concatenatedValues[0]) < 100):
          for idx2, values in enumerate(concatenatedValues):
            tabAx[int(idx/nbColumns), idx%nbColumns].plot(np.random.normal(idx2+1, 0.005*len(concatenatedValues), size=len(values)), values, 'b.', alpha=0.3, c=color[idx2] if idx2 < len(color) else 'b')
        tabAx[int(idx/nbColumns), idx%nbColumns].set_xticklabels(labels)
    plt.savefig(os.path.join(outputFolderResult, 'globalParametersInsideCategories_' + str(i+1) + '.png'))
  
  if saveDataPlottedInJson:
    outputFile = open(os.path.join(outputFolderResult, 'dataPlotted.txt'), 'w')
    outputFile.write(json.dumps(dataPlotted))
    outputFile.close()


import numpy as np
import cv2
import cvui
from zebrazoom.code.trackingFolder.trackingFunctions import calculateAngle
from zebrazoom.code.trackingFolder.trackingFunctions import distBetweenThetas
from zebrazoom.code.trackingFolder.trackingFunctions import assignValueIfBetweenRange
import math
from scipy.interpolate import UnivariateSpline
from numpy import linspace
import os.path
import csv
from zebrazoom.code.getImage.headEmbededFrame import headEmbededFrame

def getAccentuateFrameForManualPointSelect(image, hyperparameters):
  if hyperparameters["accentuateFrameForManualTailExtremityFind"]:
    frame = image.copy()
    quartileChose = 0.01
    lowVal  = int(np.quantile(frame, quartileChose))
    highVal = int(np.quantile(frame, 1 - quartileChose))
    frame[frame < lowVal]  = lowVal
    frame[frame > highVal] = highVal
    frame = frame - lowVal
    mult  = np.max(frame)
    frame = frame * (255/mult)
    frame = frame.astype(int)
    frame = (frame / np.linalg.norm(frame))*255
    return frame
  else:
    return image

def findTailTipByUserInput(frame, frameNumber, videoPath, hyperparameters):
  
  WINDOW_NAME = "Click on tail tip"
  cvui.init(WINDOW_NAME)
  cv2.moveWindow(WINDOW_NAME, 0, 0)
  
  font = cv2.FONT_HERSHEY_SIMPLEX
  frame = cv2.rectangle(frame, (0, 0), (250, 29), (255, 255, 255), -1)
  cv2.putText(frame,'Click any key if the tail is', (1, 10), font, 0.5, (0, 150, 0), 1, cv2.LINE_AA)
  cv2.putText(frame,'not straight on this image', (1, 22), font, 0.5, (0, 150, 0), 1, cv2.LINE_AA)
  
  cvui.imshow(WINDOW_NAME, frame)
  plus = 1
  while not(cvui.mouse(WINDOW_NAME, cvui.CLICK)):
    cursor = cvui.mouse(WINDOW_NAME)
    if cv2.waitKey(20) != -1:
      [frame, thresh1] = headEmbededFrame(videoPath, frameNumber + plus, hyperparameters)
      frame = cv2.rectangle(frame, (0, 0), (250, 29), (255, 255, 255), -1)
      cv2.putText(frame,'Click any key if the tail is', (1, 10), font, 0.5, (0, 150, 0), 1, cv2.LINE_AA)
      cv2.putText(frame,'not straight on this image', (1, 22), font, 0.5, (0, 150, 0), 1, cv2.LINE_AA)
      cvui.imshow(WINDOW_NAME, frame)
      plus = plus + 1
  
  cv2.destroyWindow(WINDOW_NAME)
  return [cursor.x, cursor.y]
  
def findHeadPositionByUserInput(frame, frameNumber, videoPath):
  
  WINDOW_NAME = "Click on the base of the tail"
  cvui.init(WINDOW_NAME)
  cv2.moveWindow(WINDOW_NAME, 0,0)
  font = cv2.FONT_HERSHEY_SIMPLEX
  plus = 1
  frame = cv2.rectangle(frame, (0, 0), (250, 29), (255, 255, 255), -1) 
  cv2.putText(frame,'Click any key if the tail is', (1, 10), font, 0.5, (0, 150, 0), 1, cv2.LINE_AA)
  cv2.putText(frame,'not straight on this image', (1, 22), font, 0.5, (0, 150, 0), 1, cv2.LINE_AA)
  cvui.imshow(WINDOW_NAME, frame)
  while not(cvui.mouse(WINDOW_NAME, cvui.CLICK)):
    cursor = cvui.mouse(WINDOW_NAME)
    if cv2.waitKey(20) != -1:
      [frame, thresh1] = headEmbededFrame(videoPath, frameNumber + plus, hyperparameters)
      frame = cv2.rectangle(frame, (0, 0), (250, 29), (255, 255, 255), -1)
      cv2.putText(frame,'Click any key if the tail is', (1, 10), font, 0.5, (0, 150, 0), 1, cv2.LINE_AA)
      cv2.putText(frame,'not straight on this image', (1, 22), font, 0.5, (0, 150, 0), 1, cv2.LINE_AA)
      cvui.imshow(WINDOW_NAME, frame)
      plus = plus + 1
  
  cv2.destroyWindow(WINDOW_NAME)
  return [cursor.x, cursor.y]

def getTailTipByFileSaved(hyperparameters,videoPath):
  ix = -1
  iy = -1
  with open(videoPath+'.csv') as csv_file:
    csv_reader = csv.reader(csv_file, delimiter=',')
    line_count = 0
    for row in csv_reader:
      if len(row):
        ix = row[0]
        iy = row[1]
  return [int(ix),int(iy)]

def getHeadPositionByFileSaved(videoPath):
  ix = -1
  iy = -1
  with open(videoPath+'HP.csv') as csv_file:
    csv_reader = csv.reader(csv_file, delimiter=',')
    line_count = 0
    for row in csv_reader:
      if len(row):
        ix = row[0]
        iy = row[1]
  return [int(ix),int(iy)]
  